SUBROUTINE MB02UT(TRANS, N, A, LDA, RHS, IPIV, JPIV, SCALE )
!
!     PURPOSE
!
!     To solve for x in op(A) * x = scale * RHS, using the LU factorization
!     of the N-by-N matrix A computed by SLICOT Library routine MB02UV.
!     The factorization has the form A = P * L * U * Q, where P and Q
!     are permutation matrices, L is unit lower triangular and U is
!     upper triangular. The routine is an extension to MB02UU derived 
!     from DGESC2 in LAPACK.
!
!     ARGUMENTS
!
!     Input/Output Parameters
!
!     TRANS   (input) CHARACTER 
!             If TRANS = 'N' A * x = RHS is solved or if 
!             TRANS = 'T' A**T * X = RHS is solved. 
!
!     N       (input) INTEGER
!             The order of the matrix A.
!
!     A       (input) DOUBLE PRECISION array, dimension (LDA, N)
!             The leading N-by-N part of this array must contain
!             the LU part of the factorization of the matrix A computed
!             by SLICOT Library routine MB02UV:  A = P * L * U * Q.
!
!     LDA     INTEGER
!             The leading dimension of the array A.  LDA >= max(1, N).
!
!     RHS     (input/output) DOUBLE PRECISION array, dimension (N)
!             On entry, this array must contain the right hand side
!             of the system.
!             On exit, this array contains the solution of the system.
!
!     IPIV    (input) INTEGER array, dimension (N)
!             The pivot indices; for 1 <= i <= N, row i of the
!             matrix has been interchanged with row IPIV(i).
!
!     JPIV    (input) INTEGER array, dimension (N)
!             The pivot indices; for 1 <= j <= N, column j of the
!             matrix has been interchanged with column JPIV(j).
!
!     SCALE   (output) DOUBLE PRECISION
!             The scale factor, chosen 0 < SCALE <= 1 to prevent
!             overflow in the solution.
!
!     FURTHER COMMENTS
!
!     In the interest of speed, this routine does not check the input
!     for errors. It should only be used if the order of the matrix A
!     is very small.
!
!     CONTRIBUTOR
!
!     Bo Kagstrom and P. Poromaa, Univ. of Umea, Sweden, Nov. 1993.
!     Extention to solve with op(A), Martin Koehler MPI Magdeburg
!
!     REVISIONS
!
!     April 1998 (T. Penzl).
!     Sep. 1998 (V. Sima).
!
!     ******************************************************************
!
!     .. Scalar Arguments ..
      CHARACTER          TRANS
      INTEGER            LDA, N
      DOUBLE PRECISION   SCALE
!     ..
!     .. Array Arguments ..
      INTEGER            IPIV( * ), JPIV( * )
      DOUBLE PRECISION   A( LDA, * ), RHS( * )
!     ..
!
!  =====================================================================
!
!     .. Parameters ..
      DOUBLE PRECISION   ONE, TWO
      PARAMETER          ( ONE = 1.0D+0, TWO = 2.0D+0 )
!     ..
!     .. Local Scalars ..
      INTEGER            I, J
      DOUBLE PRECISION   BIGNUM, EPS, SMLNUM, TEMP
      LOGICAL NOTRAN

!     ..
!     .. External Subroutines ..
      EXTERNAL           DLASWP, DSCAL
!     ..
!     .. External Functions ..
      INTEGER            IDAMAX
      DOUBLE PRECISION   DLAMCH
      LOGICAL            LSAME
      EXTERNAL           IDAMAX, DLAMCH, LSAME

!     ..
!     .. Intrinsic Functions ..
      INTRINSIC          ABS
!     ..
!     .. Executable Statements ..
!
!      Set constant to control owerflow
!
      EPS = DLAMCH( 'P' )
      SMLNUM = DLAMCH( 'S' ) / EPS
      BIGNUM = ONE / SMLNUM
      CALL DLABAD( SMLNUM, BIGNUM )
      NOTRAN = LSAME(TRANS,'N')

      IF ( NOTRAN ) THEN 
        !
        !     Apply permutations IPIV to RHS
        !
              CALL DLASWP( 1, RHS, LDA, 1, N-1, IPIV, 1 )
        !
        !     Solve for L part
        !
              DO 20 I = 1, N - 1
                 DO 10 J = I + 1, N
                    RHS( J ) = RHS( J ) - A( J, I )*RHS( I )
           10    CONTINUE
           20 CONTINUE
        !
        !     Solve for U part
        !
              SCALE = ONE
        !
        !     Check for scaling
        !
              I = IDAMAX( N, RHS, 1 )
              IF( TWO*SMLNUM*ABS( RHS( I ) ).GT.ABS( A( N, N ) ) ) THEN
                 TEMP = ( ONE / TWO ) / ABS( RHS( I ) )
                 CALL DSCAL( N, TEMP, RHS( 1 ), 1 )
                 SCALE = SCALE*TEMP
              END IF
        !
              DO 40 I = N, 1, -1
                 TEMP = ONE / A( I, I )
                 RHS( I ) = RHS( I )*TEMP
                 DO 30 J = I + 1, N
                    RHS( I ) = RHS( I ) - RHS( J )*( A( I, J )*TEMP )
           30    CONTINUE
           40 CONTINUE
        !
        !     Apply permutations JPIV to the solution (RHS)
        !
              CALL DLASWP( 1, RHS, LDA, 1, N-1, JPIV, -1 )
      ELSE 
        !
        !     Apply permutations JPIV to RHS
        !
              CALL DLASWP( 1, RHS, LDA, 1, N-1, JPIV, 1 )
        !
        !     Solve for U part
        !
              SCALE = ONE
        !
        !     Check for scaling
        !
              I = IDAMAX( N, RHS, 1 )
              IF( TWO*SMLNUM*ABS( RHS( I ) ).GT.ABS( A( N, N ) ) ) THEN
                 TEMP = ( ONE / TWO ) / ABS( RHS( I ) )
                 CALL DSCAL( N, TEMP, RHS( 1 ), 1 )
                 SCALE = SCALE*TEMP
              END IF
        !
              DO 140 I = 1, N
                 TEMP = ONE / A( I, I )
                 RHS( I ) = RHS( I )*TEMP
                 DO 130 J = I +1 , N
                    RHS( J ) = RHS( J ) - RHS( I  )* A( I, J )
           130    CONTINUE
           140 CONTINUE
        !
        !     Solve for L part
        !
              DO 120 I = N,1,-1
                 DO 110 J = I-1, 1, -1
                    RHS( J ) = RHS( J ) - A( I , J )*RHS( I )
           110    CONTINUE
           120 CONTINUE
        !
        !     Apply permutations JPIV to the solution (RHS)
        !
              CALL DLASWP( 1, RHS, LDA, 1, N-1, IPIV, -1 )
      END IF 

      RETURN
!
!     End of MB02UT
!
      END
